from collections import defaultdict
from io import StringIO, BytesIO


def nested_defaultdict():
    return defaultdict(nested_defaultdict)


class ContextualStringIO(StringIO):
    async def __aenter__(self):
        return self

    async def __aexit__(self, *args):
        self.close()
        return False


class ContextualBytesIO(BytesIO):
    async def __aenter__(self):
        return self

    async def __aexit__(self, *args):
        self.close()
        return False



from aiofm.protocols.s3 import S3Protocol


def main():
    protocol = S3Protocol()

    for path in protocol.ls('/rtu-datasets/own_transport/'):
        if str(path) == '/rtu-datasets/own_transport':
            continue

        print(path)

        with protocol.open(path, 'rb') as fi, protocol.open(f'{path}.copy', 'wb') as fo:
            for chunk in fi:
                fo.write(chunk)

        with protocol.open(path, 'rb') as fi, protocol.open(f'{path}.copy', 'wb') as fo:
            fo.write(fi)


if __name__ == '__main__':
    main()



from typing import Dict, Generator

import pytest


@pytest.fixture(scope='session')
def fs_tree() -> Dict[str, Dict]:
    return {
        'tmp': {
            'existing_empty_dir': {},
            'existing_dir': {
                'another_existing.txt': b'another data'
            },
            'existing.txt': b'data data data',
        },
    }


@pytest.fixture(scope='session')
def fs_list(fs_tree) -> Dict:
    def flatten(root) -> Generator[tuple, None, None]:
        for k, v in root.items():
            if type(v) is dict:
                for sub_k, sub_v in flatten(v):
                    yield f'{k.rstrip("/")}/{sub_k}', sub_v
            else:
                yield k, v

    return dict(flatten(fs_tree))


@pytest.fixture(scope='function')
def s3_resource(mocker, fs_list):
    bucket_objects = []

    for key in fs_list:
        obj = mocker.MagicMock()
        obj.key = key
        bucket_objects.append(obj)

    resource = mocker.MagicMock()
    bucket = mocker.MagicMock()
    bucket.objects.filter.return_value = bucket_objects
    resource.Bucket.return_value = bucket

    return resource



import collections.abc
import logging
from contextlib import asynccontextmanager
from io import BytesIO
from pathlib import PurePath
from typing import Sequence, Tuple, Generator

import boto3
import urllib3
from botocore.response import StreamingBody
from minio import Minio
from pydantic import SecretStr

from aiofm.protocols import BaseProtocol

logger = logging.getLogger(__name__)


class S3ReadableFile:
    def __init__(self, stream):
        self.stream = stream

    def read(self, size=-1):
        if size == -1:
            for chunk in self.stream.iter_chunks(16777216):  # 16MB
                yield chunk
        else:
            for chunk in self.stream.iter_chunks(size):
                yield chunk

    def __iter__(self):
        yield from self.read()

    def close(self):
        self.stream.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class S3WritableFile(BytesIO):
    def __init__(self, bucket_name: str, object_key: str, s3_client):
        super().__init__()
        self.bucket_name = bucket_name
        self.object_key = object_key
        self.s3_client = s3_client

    def write(self, data):
        if isinstance(data, StreamingBody):
            self.s3_client.upload_fileobj(data, self.bucket_name, self.object_key)
        elif isinstance(data, S3ReadableFile):
            self.s3_client.upload_fileobj(data.stream, self.bucket_name, self.object_key)
        elif isinstance(data, bytes):
            super().write(data)
        else:
            raise ValueError(f'Unsupported data type: {type(data)}')

    def close(self):
        self.flush()
        self.seek(0)
        self.s3_client.upload_fileobj(self, Bucket=self.bucket_name, Key=self.object_key)
        super().close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class S3Protocol(BaseProtocol):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.client = None
        self.__init_resource(**kwargs)

    def __init_resource(self, **kwargs):
        try:
            params = kwargs.get('params')

            if params:
                auth = params.get('auth')
                self.client = boto3.client(
                    's3',
                    region_name=None,
                    use_ssl=params['USE_SSL'],
                    verify=params['VERIFY_SSL_CERTIFICATES'],
                    endpoint_url=params['S3_ENDPOINT_URL'],
                    aws_access_key_id=auth['access_key'],
                    aws_secret_access_key=auth['access_secret'],
                )
            else:
                self.client = boto3.client('s3')
        except (TypeError, KeyError):
            logger.warning('No auth params has been provided for S3')

            self.client = None

    @staticmethod
    def _split_path(path: str | PurePath) -> Sequence[str]:
        try:
            path_parts = path.parts
        except AttributeError:
            path_parts = PurePath(path).parts

        if path_parts[0] == '/':
            bucket_name, prefix_parts = path_parts[1], path_parts[2:]
        else:
            bucket_name, prefix_parts = path_parts[0], path_parts[1:]

        if prefix_parts:
            prefix = '/'.join(prefix_parts)
        else:
            prefix = ''

        return bucket_name, prefix

    def ls(self, path: str | PurePath, pattern: str = None, *args, **kwargs) -> Generator[PurePath, None, None]:
        bucket_name, prefix = self._split_path(path)

        paginator = self.client.get_paginator('list_objects')
        page_iterator = paginator.paginate(Bucket=bucket_name, Prefix=prefix)

        for page in page_iterator:
            for item in page['Contents']:
                yield PurePath(f'/{bucket_name}/{item["Key"]}')

    def open(self, path: str | PurePath, *args, **kwargs):
        mode = kwargs.pop('mode', args[0] if len(args) else 'r')
        bucket_name, path = self._split_path(path)

        try:
            if 'b' not in mode:
                raise ValueError('S3 files must be opened in binary mode')

            if '+' in mode:
                raise ValueError('S3 files do not support "+" mode')

            mode = mode.replace('b', '')

            if mode not in {'r', 'w'}:
                raise ValueError(f'Invalid mode: {mode}')

            if mode == 'r':
                obj = self.client.get_object(Bucket=bucket_name, Key=path)
                stream = StreamingBody(raw_stream=obj['Body'], content_length=obj['ContentLength'])

                return S3ReadableFile(stream)
            elif mode == 'w':
                return S3WritableFile(bucket_name, path, self.client)
        except FileNotFoundError:
            if 'w' in mode or 'a' in mode:
                return S3WritableFile(bucket_name, path, self.client)
            else:
                raise

    async def exists(self, path: str | PurePath) -> bool:
        try:
            self._get_tree_item(self.tree, path)
        except FileNotFoundError:
            return False

        return True

    async def cp(self, src_path: str | PurePath, dst_path: str | PurePath):
        src_path = PurePath(src_path)
        dst_path_is_dir = isinstance(dst_path, str) and (dst_path.endswith('/') or dst_path.endswith('\\'))
        dst_path = PurePath(dst_path)
        src_item = self._get_tree_item(self.tree, src_path)
        src_is_dir = isinstance(src_item, collections.abc.Mapping)

        try:
            dst_item = self._get_tree_item(self.tree, dst_path)

            if isinstance(dst_item, collections.abc.Mapping):
                if src_is_dir:
                    pass
                else:
                    self._set_tree_item(self.tree, dst_path.joinpath(src_path.name), src_item)
            else:
                if dst_path_is_dir:
                    raise ValueError(f'Unable to copy {src_path} to directory path. it is a file')

                if src_is_dir:
                    raise ValueError(f'Unable to copy directory {src_path} to file {dst_path}')
                else:
                    self._set_tree_item(self.tree, dst_path, src_item)

                self._set_tree_item(self.tree, dst_path, src_item)
        except FileNotFoundError:
            if src_is_dir:
                pass
            else:
                if dst_path_is_dir:
                    self._set_tree_item(self.tree, dst_path.joinpath(src_path.name), src_item)
                else:
                    self._set_tree_item(self.tree, dst_path, src_item)

    async def mkdir(self, path: str | PurePath):
        return

    async def mkdirs(self, path: str | PurePath):
        return

    async def mv(self, src_path: str | PurePath, dst_path: str | PurePath):
        await self.cp(src_path, dst_path)
        await self.rm(src_path)

    async def rm(self, path: str | PurePath):
        """
        Removes file
        """

        if await self.is_dir(path):
            bucket_name, path = self._split_path(path)
            key_list = self.resource.Bucket(bucket_name).objects.filter(Prefix=f'{path}/')

            for key in key_list:
                self.resource.Object(bucket_name, key.key).delete()
        else:
            bucket_name, path = self._split_path(path)
            self.resource.Object(bucket_name, path).delete()

    async def is_dir(self, path: str | PurePath) -> bool:
        return isinstance(self._get_tree_item(self.tree, path), collections.abc.Mapping)

    async def glob(self, pattern: str) -> Tuple:
        pass


def _get_minio_client(endpoint_url: str, region_name: str, access_key_id: SecretStr, secret_access_key: SecretStr,
                      secure: bool = True) -> Minio:
    http_client = urllib3.PoolManager(
        timeout=urllib3.util.Timeout(connect=300, read=300),
        maxsize=10,
        cert_reqs='CERT_REQUIRED' if secure else 'CERT_NONE',
        retries=urllib3.Retry(
            total=5,
            backoff_factor=0.2,
            status_forcelist=[500, 502, 503, 504]
        )
    )

    client = Minio(
        endpoint_url,
        region=region_name,
        access_key=access_key_id.get_secret_value(),
        secret_key=secret_access_key.get_secret_value(),
        http_client=http_client,
        secure=secure
    )

    return client


class MinioProtocol(BaseProtocol):
    def __init__(self, endpoint_url: str, region_name: str, access_key_id: SecretStr, secret_access_key: SecretStr,
                 secure: bool = True, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.client: Minio = _get_minio_client(endpoint_url, region_name, access_key_id, secret_access_key, secure)

    @staticmethod
    def _split_path(path: str | PurePath):
        bucket_name, *parts = super()._split_path(path)
        bucket_path = str(PurePath('/', *parts))

        return bucket_name, bucket_path

    def ls(self, path: str, pattern: str = None, *args, **kwargs) -> Sequence:
        recursive = bool(kwargs.get('recursive'))
        bucket_name, prefix = self._split_path(path)

        objects = self.client.list_objects(bucket_name, prefix, recursive)

        return tuple(objects)

    @asynccontextmanager
    def open(self, path: str, *args, **kwargs):
        raise NotImplemented

    def exists(self, path: str) -> bool:
        raise NotImplemented

    def cp(self, src_path: str, dst_path: str):
        raise NotImplemented

    def mkdir(self, path: str):
        raise NotImplemented

    def mkdirs(self, path: str):
        raise NotImplemented

    def mv(self, src_path: str, dst_path: str):
        raise NotImplemented

    def rm(self, path: str | PurePath):
        """
        Removes file
        """

        if self.is_dir(path):
            bucket_name, path = self._split_path(path)
            key_list = self.resource.Bucket(bucket_name).objects.filter(Prefix=f'{path}/')

            for key in key_list:
                self.resource.Object(bucket_name, key.key).delete()
        else:
            bucket_name, path = self._split_path(path)
            self.resource.Object(bucket_name, path).delete()

    def is_dir(self, path: str) -> bool:
        raise NotImplemented

    def glob(self, pattern: str) -> Tuple:
        raise NotImplemented



from abc import ABCMeta, abstractmethod
from pathlib import PurePath
from typing import Sequence, Tuple


class BaseProtocol(metaclass=ABCMeta):
    def __init__(self, *args, **kwargs):
        super().__init__()

    @staticmethod
    def _split_path(path: str | PurePath):
        try:
            path_parts = path.parts
        except AttributeError:
            path_parts = PurePath(path).parts

        return path_parts

    @staticmethod
    @abstractmethod
    async def ls(path: str, pattern: str = None, *args, **kwargs) -> Sequence:
        pass

    @abstractmethod
    async def open(self, path, *args, **kwargs):
        pass

    @staticmethod
    @abstractmethod
    async def exists(path: str) -> bool:
        pass

    @staticmethod
    @abstractmethod
    async def cp(src_path: str, dst_path: str):
        pass

    @abstractmethod
    async def mkdir(self, path: str):
        pass

    @abstractmethod
    async def mkdirs(self, path: str):
        pass

    @abstractmethod
    async def mv(self, src_path: str, dst_path: str):
        pass

    @abstractmethod
    async def rm(self, path: str):
        pass

    @staticmethod
    @abstractmethod
    async def is_dir(path: str) -> bool:
        pass

    @abstractmethod
    async def glob(self, pattern: str) -> Tuple:
        pass


def get_protocol_for_path(path: str) -> BaseProtocol:
    raise NotImplemented



import collections.abc
import operator
from contextlib import asynccontextmanager
from functools import reduce
from pathlib import PurePath
from typing import Any, Mapping, Sequence, Tuple

from aiofm.helpers import ContextualBytesIO, ContextualStringIO
from aiofm.protocols import BaseProtocol


class MemoryProtocol(BaseProtocol):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.tree = {}

    @staticmethod
    def _split_path(path: str | PurePath):
        try:
            path_parts = path.parts
        except AttributeError:
            path_parts = PurePath(path).parts

        return path_parts

    @classmethod
    def _remove_tree_item(cls, tree: Mapping, path: str | PurePath):
        path_parts = cls._split_path(path)

        try:
            del reduce(operator.getitem, path_parts[:-1], tree)[path_parts[-1]]
        except KeyError:
            pass

    @classmethod
    def _get_tree_item(cls, tree: Mapping, path: str | PurePath):
        path_parts = cls._split_path(path)

        try:
            return reduce(operator.getitem, path_parts, tree)
        except KeyError:
            raise FileNotFoundError

    @classmethod
    def _set_tree_item(cls, tree: Mapping, path: str | PurePath, value: Any):
        path_parts = cls._split_path(path)
        current_node = tree

        for path_part in path_parts[:-1]:
            try:
                current_node = current_node[path_part]

                if not isinstance(current_node, collections.abc.Mapping):
                    raise FileNotFoundError(f'Node already exists: {current_node}')
            except KeyError:
                new_node = {}
                current_node[path_part] = new_node
                current_node = new_node

        try:
            if type(current_node[path_parts[-1]]) != type(value):
                raise FileNotFoundError('Node already exists')

            current_node[path_parts[-1]] = value
        except KeyError:
            current_node[path_parts[-1]] = value

    @classmethod
    def _set_parent_tree_item(cls, tree: Mapping, path: str | PurePath, value: Any):
        try:
            parent_item_path = path.parent
        except AttributeError:
            parent_item_path = PurePath(path).parent

        cls._set_tree_item(tree, parent_item_path, value)

    async def ls(self, path: str | PurePath, pattern: str = None, *args, **kwargs) -> Sequence:
        item = self._get_tree_item(self.tree, path)

        return tuple(item)

    @asynccontextmanager
    async def open(self, path: str | PurePath, *args, **kwargs):
        mode = kwargs.pop('mode', args[0] if len(args) else 'r')

        try:
            encoding = kwargs.get('encoding', 'utf-8')
            item = self._get_tree_item(self.tree, path)

            if 'b' in mode:
                f = ContextualBytesIO(item)
            else:
                f = ContextualStringIO(item.decode(encoding))

            yield f

            if 'w' in mode:
                if 'b' in mode:
                    item = f.getvalue()
                else:
                    item = f.getvalue().encode(encoding)

                self._set_tree_item(self.tree, path, item)
        except FileNotFoundError:
            if 'w' in mode or 'a' in mode:
                f = ContextualStringIO(None, *args, **kwargs)

                yield f

                self._set_tree_item(self.tree, path, f)
            else:
                raise

    async def exists(self, path: str | PurePath) -> bool:
        try:
            self._get_tree_item(self.tree, path)
        except FileNotFoundError:
            return False

        return True

    async def cp(self, src_path: str | PurePath, dst_path: str | PurePath):
        src_path = PurePath(src_path)
        dst_path_is_dir = isinstance(dst_path, str) and (dst_path.endswith('/') or dst_path.endswith('\\'))
        dst_path = PurePath(dst_path)
        src_item = self._get_tree_item(self.tree, src_path)
        src_is_dir = isinstance(src_item, collections.abc.Mapping)

        try:
            dst_item = self._get_tree_item(self.tree, dst_path)

            if isinstance(dst_item, collections.abc.Mapping):
                if src_is_dir:
                    pass
                else:
                    self._set_tree_item(self.tree, dst_path.joinpath(src_path.name), src_item)
            else:
                if dst_path_is_dir:
                    raise ValueError(f'Unable to copy {src_path} to directory path. it is a file')

                if src_is_dir:
                    raise ValueError(f'Unable to copy directory {src_path} to file {dst_path}')
                else:
                    self._set_tree_item(self.tree, dst_path, src_item)

                self._set_tree_item(self.tree, dst_path, src_item)
        except FileNotFoundError:
            if src_is_dir:
                pass
            else:
                if dst_path_is_dir:
                    self._set_tree_item(self.tree, dst_path.joinpath(src_path.name), src_item)
                else:
                    self._set_tree_item(self.tree, dst_path, src_item)

    async def mkdir(self, path: str | PurePath):
        await self.mkdirs(path)

    async def mkdirs(self, path: str | PurePath):
        self._set_tree_item(self.tree, path, {})

    async def mv(self, src_path: str | PurePath, dst_path: str | PurePath):
        await self.cp(src_path, dst_path)
        await self.rm(src_path)

    async def rm(self, path: str | PurePath):
        self._remove_tree_item(self.tree, path)

    async def is_dir(self, path: str | PurePath) -> bool:
        return isinstance(self._get_tree_item(self.tree, path), collections.abc.Mapping)

    async def glob(self, pattern: str) -> Tuple:
        pass



from contextlib import asynccontextmanager


class FileManager:
    @staticmethod
    def cp(src_path: str, dst_path: str):
        raise NotImplemented

    @staticmethod
    def rm(path: str):
        raise NotImplemented

    @staticmethod
    def ls(path: str):
        raise NotImplemented

    @staticmethod
    def mkdir(path: str):
        raise NotImplemented

    @staticmethod
    @asynccontextmanager
    def open(path: str, mode: str = 'r'):
        raise NotImplemented



import pytest

from aiofm.protocols.memory import MemoryProtocol


@pytest.mark.asyncio
async def test_ls_tmp_dir():
    fs = MemoryProtocol()
    fs.tree = {'/': {'tmp': {'xxx': {}, 'a.txt': b'data data data'}}}

    assert await fs.ls('/tmp') == ('xxx', 'a.txt')


@pytest.mark.asyncio
async def test_ls_inextisting_dir_fails():
    fs = MemoryProtocol()
    fs.tree = {'/': {'tmp': {'xxx': {}, 'a.txt': b'data data data'}}}

    with pytest.raises(FileNotFoundError):
        assert await fs.ls('/pmt')


@pytest.mark.asyncio
async def test_existing_file_exists():
    fs = MemoryProtocol()
    fs.tree = {'/': {'tmp': {'xxx': {}, 'a.txt': b'data data data'}}}

    assert await fs.exists('/tmp/a.txt') is True


@pytest.mark.asyncio
async def test_inexisting_file_does_not_exist():
    fs = MemoryProtocol()
    fs.tree = {'/': {'tmp': {'xxx': {}, 'a.txt': b'data data data'}}}

    assert await fs.exists('/tmp/b.txt') is False


@pytest.mark.asyncio
async def test_existing_dir_exists():
    fs = MemoryProtocol()
    fs.tree = {'/': {'tmp': {'xxx': {}, 'a.txt': b'data data data'}}}

    assert await fs.exists('/tmp/xxx') is True


@pytest.mark.asyncio
async def test_inexisting_dir_does_not_exist():
    fs = MemoryProtocol()
    fs.tree = {'/': {'tmp': {'xxx': {}, 'a.txt': b'data data data'}}}

    assert await fs.exists('/tmp/yyy') is False


@pytest.mark.asyncio
async def test_existing_path_exists():
    fs = MemoryProtocol()
    fs.tree = {'/': {'tmp': {'xxx': {}, 'a.txt': b'data data data'}}}

    assert await fs.exists('/tmp/a.txt') is True


@pytest.mark.asyncio
async def test_inexisting_path_does_not_exist():
    fs = MemoryProtocol()
    fs.tree = {'/': {'tmp': {'xxx': {}, 'a.txt': b'data data data'}}}

    assert await fs.exists('/pmt/a.txt') is False


@pytest.mark.asyncio
async def test_file_is_not_a_directory():
    fs = MemoryProtocol()
    fs.tree = {'/': {'tmp': {'xxx': {}, 'a.txt': b'data data data'}}}

    assert await fs.is_dir('/tmp/a.txt') is False


@pytest.mark.asyncio
async def test_directory_is_a_direcotry():
    fs = MemoryProtocol()
    fs.tree = {'/': {'tmp': {'xxx': {}, 'a.txt': b'data data data'}}}

    assert await fs.is_dir('/tmp/xxx') is True


@pytest.mark.asyncio
async def test_inexisting_path_is_dir_check_fails():
    fs = MemoryProtocol()
    fs.tree = {'/': {'tmp': {'xxx': {}, 'a.txt': b'data data data'}}}

    with pytest.raises(FileNotFoundError):
        assert await fs.is_dir('/tmp/yyy.txt')


@pytest.mark.asyncio
async def test_mkdir_first_toplevel():
    fs = MemoryProtocol()

    await fs.mkdir('/home')

    assert fs.tree == {'/': {'home': {}}}


@pytest.mark.asyncio
async def test_mkdir_nested():
    fs = MemoryProtocol()
    fs.tree = {'/': {'home': {}}}

    await fs.mkdir('/home/user')

    assert fs.tree == {'/': {'home': {'user': {}}}}


@pytest.mark.asyncio
async def test_mkdir_existing_dir():
    fs = MemoryProtocol()
    fs.tree = {'/': {'home': {'user': {}}}}

    await fs.mkdir('/home/user')

    assert fs.tree == {'/': {'home': {'user': {}}}}


@pytest.mark.asyncio
async def test_mkdir_nested_within_inexisting_path():
    fs = MemoryProtocol()
    fs.tree = {'/': {}}

    await fs.mkdir('/home/user/documents')

    assert fs.tree == {'/': {'home': {'user': {'documents': {}}}}}


@pytest.mark.asyncio
async def test_mkdir_first_nested():
    fs = MemoryProtocol()

    await fs.mkdir('/home/user/documents')

    assert fs.tree == {'/': {'home': {'user': {'documents': {}}}}}


@pytest.mark.asyncio
async def test_mkdir_fails_when_path_is_a_file():
    fs = MemoryProtocol()
    fs.tree = {'/': {'tmp': {'xxx': {}, 'a.txt': b'data data data'}}}

    with pytest.raises(FileNotFoundError):
        await fs.mkdir('/tmp/a.txt')

    with pytest.raises(FileNotFoundError):
        await fs.mkdir('/tmp/a.txt/dir')


@pytest.mark.asyncio
async def test_mkdirs_first_toplevel():
    fs = MemoryProtocol()

    await fs.mkdirs('/home')

    assert fs.tree == {'/': {'home': {}}}


@pytest.mark.asyncio
async def test_mkdirs_nested():
    fs = MemoryProtocol()
    fs.tree = {'/': {'home': {}}}

    await fs.mkdirs('/home/user')

    assert fs.tree == {'/': {'home': {'user': {}}}}


@pytest.mark.asyncio
async def test_mkdirs_existing_dir():
    fs = MemoryProtocol()
    fs.tree = {'/': {'home': {'user': {}}}}

    await fs.mkdirs('/home/user')

    assert fs.tree == {'/': {'home': {'user': {}}}}


@pytest.mark.asyncio
async def test_mkdirs_nested_within_inexisting_path():
    fs = MemoryProtocol()
    fs.tree = {'/': {}}

    await fs.mkdirs('/home/user/documents')

    assert fs.tree == {'/': {'home': {'user': {'documents': {}}}}}


@pytest.mark.asyncio
async def test_mkdirs_first_nested():
    fs = MemoryProtocol()

    await fs.mkdirs('/home/user/documents')

    assert fs.tree == {'/': {'home': {'user': {'documents': {}}}}}


@pytest.mark.asyncio
async def test_mkdirs_fails_when_path_is_a_file():
    fs = MemoryProtocol()
    fs.tree = {'/': {'tmp': {'xxx': {}, 'a.txt': b'data data data'}}}

    with pytest.raises(FileNotFoundError):
        await fs.mkdirs('/tmp/a.txt')

    with pytest.raises(FileNotFoundError):
        await fs.mkdirs('/tmp/a.txt/dir')


@pytest.mark.asyncio
async def test_rm_empty_dir():
    fs = MemoryProtocol()
    fs.tree = {'/': {'home': {'user': {'documents': {}}}}}

    await fs.rm('/home/user/documents')

    assert fs.tree == {'/': {'home': {'user': {}}}}


@pytest.mark.asyncio
async def test_rm_non_empty_dir():
    fs = MemoryProtocol()
    fs.tree = {'/': {'tmp': {'xxx': {}, 'a.txt': b'data data data'}}}

    await fs.rm('/tmp')

    assert fs.tree == {'/': {}}


@pytest.mark.asyncio
async def test_rm_file():
    fs = MemoryProtocol()
    fs.tree = {'/': {'tmp': {'xxx': {}, 'a.txt': b'data data data'}}}

    await fs.rm('/tmp/a.txt')

    assert fs.tree == {'/': {'tmp': {'xxx': {}}}}


@pytest.mark.asyncio
async def test_rm_inexisting_dir():
    fs = MemoryProtocol()

    await fs.rm('/home/user/documents')

    assert fs.tree == {}

    fs.tree = {'/': {'home': {'user': {'documents': {}}}}}

    await fs.rm('/tmp')

    assert fs.tree == {'/': {'home': {'user': {'documents': {}}}}}


@pytest.mark.asyncio
async def test_mv_file_to_existing_dir():
    fs = MemoryProtocol()
    fs.tree = {'/': {'tmp': {'xxx': {}, 'a.txt': b'data data data'}}}

    await fs.mv('/tmp/a.txt', '/tmp/xxx')

    assert fs.tree == {'/': {'tmp': {'xxx': {'a.txt': b'data data data'}}}}


@pytest.mark.asyncio
async def test_mv_file_to_inexisting_dir():
    fs = MemoryProtocol()
    fs.tree = {'/': {'tmp': {'xxx': {}, 'a.txt': b'data data data'}}}

    await fs.mv('/tmp/a.txt', '/tmp/yyy/')

    assert fs.tree == {'/': {'tmp': {'xxx': {}, 'yyy': {'a.txt': b'data data data'}}}}


@pytest.mark.asyncio
async def test_mv_file_to_existing_file():
    fs = MemoryProtocol()
    fs.tree = {'/': {'tmp': {'xxx': {'b.txt': b''}, 'a.txt': b'data data data'}}}

    await fs.mv('/tmp/a.txt', '/tmp/xxx/b.txt')

    assert fs.tree == {'/': {'tmp': {'xxx': {'b.txt': b'data data data'}}}}


@pytest.mark.asyncio
async def test_mv_file_to_inexisting_file():
    fs = MemoryProtocol()
    fs.tree = {'/': {'tmp': {'xxx': {}, 'a.txt': b'data data data'}}}

    await fs.mv('/tmp/a.txt', '/tmp/xxx/c.txt')

    assert fs.tree == {'/': {'tmp': {'xxx': {'c.txt': b'data data data'}}}}


@pytest.mark.asyncio
async def test_mv_dir_to_existing_file_should_fail():
    fs = MemoryProtocol()
    fs.tree = {'/': {'tmp': {'xxx': {}, 'a.txt': b'data data data'}}}

    with pytest.raises(ValueError, match='Unable to copy directory /tmp/xxx to file /tmp/a.txt'):
        await fs.mv('/tmp/xxx', '/tmp/a.txt')


@pytest.mark.asyncio
async def test_cp_file_to_existing_dir():
    fs = MemoryProtocol()
    fs.tree = {'/': {'tmp': {'xxx': {}, 'a.txt': b'data data data'}}}

    await fs.cp('/tmp/a.txt', '/tmp/xxx')

    assert fs.tree == {'/': {'tmp': {'a.txt': b'data data data', 'xxx': {'a.txt': b'data data data'}}}}


@pytest.mark.asyncio
async def test_cp_file_to_inexisting_dir():
    fs = MemoryProtocol()
    fs.tree = {'/': {'tmp': {'xxx': {}, 'a.txt': b'data data data'}}}

    await fs.cp('/tmp/a.txt', '/tmp/yyy/')

    assert fs.tree == {'/': {'tmp': {'a.txt': b'data data data', 'xxx': {}, 'yyy': {'a.txt': b'data data data'}}}}


@pytest.mark.asyncio
async def test_cp_file_to_existing_file():
    fs = MemoryProtocol()
    fs.tree = {'/': {'tmp': {'xxx': {'b.txt': b''}, 'a.txt': b'data data data'}}}

    await fs.cp('/tmp/a.txt', '/tmp/xxx/b.txt')

    assert fs.tree == {'/': {'tmp': {'a.txt': b'data data data', 'xxx': {'b.txt': b'data data data'}}}}


@pytest.mark.asyncio
async def test_cp_file_to_inexisting_file():
    fs = MemoryProtocol()
    fs.tree = {'/': {'tmp': {'xxx': {}, 'a.txt': b'data data data'}}}

    await fs.cp('/tmp/a.txt', '/tmp/xxx/c.txt')

    assert fs.tree == {'/': {'tmp': {'xxx': {'c.txt': b'data data data'}, 'a.txt': b'data data data'}}}


@pytest.mark.asyncio
async def test_cp_dir_to_existing_file_should_fail():
    fs = MemoryProtocol()
    fs.tree = {'/': {'tmp': {'xxx': {}, 'a.txt': b'data data data'}}}

    with pytest.raises(ValueError, match='Unable to copy directory /tmp/xxx to file /tmp/a.txt'):
        await fs.cp('/tmp/xxx', '/tmp/a.txt')


@pytest.mark.asyncio
async def test_open_inexisting_file_for_read_should_fail():
    fs = MemoryProtocol()
    fs.tree = {'/': {'tmp': {'xxx': {}, 'a.txt': b'data data data'}}}

    with pytest.raises(FileNotFoundError):
        async with fs.open('/tmp/b.txt'):
            pass


@pytest.mark.asyncio
async def test_open_existing_file_for_read():
    fs = MemoryProtocol()
    fs.tree = {'/': {'tmp': {'xxx': {}, 'a.txt': b'data data data'}}}

    async with fs.open('/tmp/a.txt') as f:
        assert f.read() == 'data data data'


@pytest.mark.asyncio
async def test_unclosed_file_does_not_change_fs():
    fs = MemoryProtocol()
    fs.tree = {'/': {'tmp': {'xxx': {}, 'a.txt': b'data data data'}}}

    async with fs.open('/tmp/a.txt', mode='w') as f:
        f.write('TEST TEST TEST')

        assert fs.tree['/']['tmp']['a.txt'] == b'data data data'


@pytest.mark.asyncio
async def test_closed_file_changes_fs():
    fs = MemoryProtocol()
    fs.tree = {'/': {'tmp': {'xxx': {}, 'a.txt': b'data data data'}}}

    async with fs.open('/tmp/a.txt', mode='w') as f:
        f.write('TEST TEST TEST')

    assert fs.tree['/']['tmp']['a.txt'] == b'TEST TEST TEST'
